#!/usr/bin/env python
import argparse

import numpy as np
import spglib

from pypolymlp.core.interface_vasp import Poscar
from pypolymlp.utils.vasp_utils import print_poscar, write_poscar_file


class SymCell:

    def __init__(self, poscar_name=None, st_dict=None, symprec=1e-4):

        if poscar_name is not None:
            st_dict = Poscar(poscar_name).get_structure()

        if "comment" in st_dict:
            self.comment = st_dict["comment"]
        else:
            self.comment = "Generated by pypolymlp and spglib"

        types = st_dict["types"]
        elements = st_dict["elements"]
        self.n_types = len(st_dict["n_atoms"])
        self.cell = (
            np.array(st_dict["axis"]).T,
            np.array(st_dict["positions"]).T,
            np.array(types),
        )

        self.element_map = dict()
        for e, t in zip(elements, types):
            self.element_map[t] = e

        self.symprec = symprec

    def refine_cell(self, standardize_cell=False):

        if standardize_cell is False:
            try:
                lattice1, position1, types1 = spglib.refine_cell(
                    self.cell, symprec=self.symprec
                )
            except:
                raise TypeError
        else:
            try:
                lattice1, position1, types1 = spglib.standardize_cell(
                    self.cell, symprec=self.symprec
                )
            except:
                raise TypeError

        type_list = list(range(0, self.n_types))

        position1_tmp, types1_tmp = [], []
        for t_ref in type_list:
            for i, t in enumerate(types1):
                if t == t_ref:
                    position1_tmp.append(position1[i])
                    types1_tmp.append(t)
        position1, types1 = np.array(position1_tmp), types1_tmp

        n_atoms1 = [types1.count(t_ref) for t_ref in type_list]
        elements1 = [self.element_map[t] for t in types1]

        st_dict = dict()
        st_dict["axis"] = lattice1.T
        st_dict["positions"] = position1.T
        st_dict["n_atoms"] = n_atoms1
        st_dict["elements"] = elements1
        st_dict["types"] = types1
        st_dict["comment"] = self.comment
        return st_dict

    def get_spacegroup(self):
        return spglib.get_spacegroup(self.cell, symprec=self.symprec)

    def get_spacegroup_multiple_prec(self, symprecs=[1e-2, 1e-3, 1e-4, 1e-5]):
        return [spglib.get_spacegroup(self.cell, symprec=p) for p in symprecs]


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--poscar", type=str, help="poscar file name")
    parser.add_argument(
        "--symprec",
        type=float,
        default=1e-4,
        help="numerical precision for finding symmetry",
    )
    parser.add_argument("--refine_cell", action="store_true", help="refine cell")
    parser.add_argument(
        "--standardize_cell", action="store_true", help="standardize cell"
    )
    parser.add_argument("--space_group", action="store_true", help="get space group")

    args = parser.parse_args()
    sc = SymCell(args.poscar, symprec=args.symprec)

    if args.refine_cell:
        st_dict = sc.refine_cell()
        print_poscar(st_dict)
        write_poscar_file(st_dict)
    elif args.standardize_cell:
        st_dict = sc.refine_cell(standardize_cell=True)
        print_poscar(st_dict)
        write_poscar_file(st_dict)

    if args.space_group:
        print(" space_group = ", sc.get_spacegroup())
